(* This is a (fairly large) subset of the syntax specified at the back
 * of the "Elements of ML Programming" by Ullman
 *)

structure MLAst = 
struct
    datatype
	prog	= prog of strdecs
    	
    and strdec 	= non_structure_dec of dec
                | structure_dec of id * struct'
	
    and struct'	= struct_dec of strdecs
                | struct_name_copy of id
                | functor_application of id * functor_app_args
    and functor_app_args = faa_struct of struct'
                         | faa_strdecs of strdecs
		
    and dec  = val_dec of pat_exp_bindings 
	     | fun_dec of fbind
	     | type_dec of tbind
             | datatype_dec of dtbind * tbind
	     | exception_dec of id_type_bindings 
	     | open_dec of ids

    and pat_exp_binding = pat_exp_binding of pat * exp

    and id_type_binding = id_type_binding of id * type'
    and tbind =	tbind of id_type_bindings

    and dtbind = dtbind of dtbinding list
    and dtbinding = dtbinding of id * (dt_constructor list)
    and dt_constructor = dt_constructor_atomic of id
                       | dt_constructor_compound of id * type'
	
    and fbind   = fbind of fun_defs
    and fun_def = fun_def of id * ids * exp

    and type'	= basetype of id | prodtype of type' * type'

    and apat	= apat_ident of id
		| apat_const of const 
		| apat_any 
		| apat_tuple of pat list
	
    and pat	= pat_atomic of apat 
		| pat_compound  of id * apat

    and match	= match of pat_exp_bindings

    and exp	= skip 
		| const of const 
		| var of id 
		| if_else of exp * exp * exp
		| tuple of exps
	  	| let_in of decs * exp 
		| case_of of exp * match
	  	| primop of primop * exp 
		| call of exp * exp 
		| anon_fun of ids * exp
		| exp_orelse of exp * exp 
		| exp_andalso of exp * exp 
		| curried_call of id * exps

    and primop	= op_neg | op_not | op_div | op_mod | op_times 
		| op_plus | op_minus | op_lt | op_leq | op_gt | op_geq 
		| op_equal | op_notequal 
		| op_andb | op_orb | op_xorb | op_lshift | op_rshift

    and const	= int of int | bool of bool | string of string

    withtype
	exps = exp list
    and id = string
    and strdecs = strdec list
    and decs = dec list
    and tbinds = tbind list
    and fun_defs = fun_def list
    and ids = id list
    and type's = type' list
    and pat_exp_bindings = pat_exp_binding list
    and id_type_bindings = id_type_binding list

end
